#ifndef __INTERPOLATOR_H__
#define __INTERPOLATOR_H__

#include <vector>
#include <map>
#include <boost/shared_ptr.hpp>
#include "bvp.hpp"
#include "linalg.hpp"
#include "func.hpp"
#include "diff_op.hpp"

namespace bvp{
  using std::map;
  using boost::shared_ptr;
  template<typename RBF>
  class interpolator : public realfunc{
  public:
    interpolator();
 
    //Interpolate given a BVP
    interpolator(shared_ptr<linear_BVP2> bvp);

    //Interpolate given some data points and the value at those points
    interpolator(const map<point, double>& Xi);

    //Same, but provide domain information. The values of Xi must
    //match points on the given domain Omega.
    interpolator(shared_ptr<domain> Omega, const map<point, double>& Xi);

    //Destroys all data already in the interpolator.
    void interpolate(const map<point, double>& Xi); 
    void interpolate(shared_ptr<linear_BVP2> bvp);
    
    //Evaluation
    double operator()(const point& p) const;
    double at(const point& p) const;

    //Derivatives
    double d(const point& p, size_t k) const;
    double d2(const point &p, size_t k1, size_t k2) const;

    //These functions allow for partial redefinition of the BVP as
    //required for the additive Schwartz domain decomposition method,
    //and for other methods.
    void set_f(const realfunc &f);
    void set_g(const realfunc &g);
    void set_f(const map<point, double>& f);
    void set_g(const map<point, double>& g);

    //Linear arithmetic operators
    interpolator<RBF> operator+(const interpolator<RBF>& u) const;
    interpolator<RBF> operator-(const interpolator<RBF>& u) const;
    interpolator<RBF> operator*(double a) const;
    interpolator<RBF> operator/(double a) const;
    
  private:
    //Once the matrix is defined, this function inverts it.
    void computecoeffs(); 
	 
    void init(shared_ptr<linear_BVP2> bvp);

    shared_ptr<linear_BVP2> thebvp;

    //Number of interior points. 
    size_t n; 
    //Number of boundary points.
    size_t m; 

    //The matrix to invert
    matrix M; 

    bool initted;
    void not_initted(int line, string file) const; //Exception thrower

    linalg::vector coeffs;
    std::vector<RBF> rbfs;
    size_t rbfs_hash;    

    size_t hash_value(const std::vector<RBF>& rbfs_in);
    
    //Data of where the interpolator has been evaluated needs a point
    //and a vector representing a multi-index of partials (an empty
    //vector represents evaluation, instead of differentiation).
    typedef std::pair<linalg::point, std::vector<size_t> > diff_data;
    
    //A remember table
    mutable map<diff_data, double> remtable;
  };

  //For comfortable syntax
  template <typename RBF>
  interpolator<RBF> operator*(double a, const interpolator<RBF>& u)
  {
    return u*a;
  }
}

#endif
