#include <vector>
#include <set>

#include <utility>

#include "include/interpolator.hpp"
#include "include/rbf.hpp"
#include "include/func.hpp"
#include "include/error.hpp"
#include "include/utils.hpp"

#include <boost/functional/hash.hpp>
#include <boost/shared_ptr.hpp>

namespace bvp{ 
  using boost::shared_ptr;

  template<typename RBF>
  interpolator<RBF>::interpolator()
  {
    initted = false;
  }

  template<typename RBF>
  interpolator<RBF>::interpolator(shared_ptr<linear_BVP2> bvp)
  {
    //Workaround because gdb can't break inside constructors. :-(
    init(bvp);
  }

  template<typename RBF>
  interpolator<RBF>::interpolator(shared_ptr<domain> Omega, 
				  const map<point, double>& Xi)
  {
    //Check that Xi matches the domain in size
    size_t Omega_size = Omega -> get_interior().size() 
                       + Omega -> get_boundary().size();
    if(Xi.size() != Omega_size)
    {
      badArgument exc;
      if(Xi.size() < Omega_size)
	exc.reason = "Did not provide enough interpolation data for every point in the given domain.";
      else
	exc.reason = "Provided more interpolation data than points in the given domain.";
      exc.line = __LINE__;
      exc.file = __FILE__;
      throw exc;
    }

    //Create a trivial bvp taking into account the given domain.
    shared_ptr<Id_op> Id(new Id_op);
    shared_ptr<dirichlet_op> D(new dirichlet_op);
    map<point, double> f, g;

    //Extract f and g from given information
    for(map<point, double>::const_iterator I = Xi.begin();
	I != Xi.end(); I++)
    {
      if(utils::contains(Omega -> get_interior(), I -> first))
	f[I -> first] = I -> second;
      else if(utils::contains(Omega -> get_boundary(), I -> first))
	g[I -> first] = I -> second;
      else
      {
	badArgument exc;
	exc.reason = "The interpolation data contains points not in the given domain.";
	exc.line = __LINE__;
	exc.file = __FILE__;
	throw exc;
      }
    } 
    
    shared_ptr<linear_BVP2> new_bvp(new linear_BVP2(Omega, Id, D, f,g));
    
    interpolate(new_bvp);

  }

  template<typename RBF>
  interpolator<RBF>::interpolator(const map<point, double>& Xi)
  {
    interpolate(Xi);
  }

  template<typename RBF>
  void interpolator<RBF>::interpolate(const map<point, double>& Xi)
  {
    
    if(Xi.empty()){//Dude, wtf?
      badArgument exc;
      exc.reason = "Cannot interpolate if no data is given.";
      exc.line = __LINE__;
      exc.file = __FILE__;
      throw exc;
    }

    //Create a trivial bvp.
    shared_ptr<Id_op> Id(new Id_op);
    shared_ptr<dirichlet_op> D(new dirichlet_op);
    set<point> intr;
    set<point> bdry; //empty
    map<point, point> nrml; //empty
    map<point, double> g; //empty
    
    bool dim_set = false;
    size_t dimension = 0;
    for(map<point,double>::const_iterator i = Xi.begin(); i != Xi.end(); i++){
      if(!dim_set){
	dimension = (i -> first).size();
	dim_set = true;
      }
      else if(dimension != (i -> first).size()){
	badArgument exc;
	exc.reason = "Inconformant dimensions in interpolation data.";
	exc.line = __LINE__;
	exc.file = __FILE__;
	throw exc;
      }
      intr.insert( i->first);
    }
    shared_ptr<domain> Omega(new domain(dimension, intr,bdry,nrml));
    shared_ptr<linear_BVP2> bvp(new linear_BVP2(Omega, Id, D, Xi, g));
    
    init(bvp);
  }

  template<typename RBF>
  void interpolator<RBF>::interpolate(shared_ptr<linear_BVP2> bvp)
  {
    init(bvp);
  }

  template<typename RBF>
  void interpolator<RBF>::init(shared_ptr<linear_BVP2> bvp)
  {
    thebvp = bvp;

    using namespace linalg;
    using std::set;
    
    shared_ptr<const domain> Omega = bvp -> get_domain();
    set<point> interior = Omega -> get_interior();
    set<point> boundary = Omega -> get_boundary();
    map<point, vector> normals = Omega -> get_normals();
    n = interior.size();
    m = boundary.size();
    
    vector temp(n+m);
    coeffs = temp;
    rbfs.reserve(n+m);

    
    RBF::set_dimension(Omega -> get_dimension());
    
    set<point>::iterator I;
    //Define all the rbfs...
    for(I = interior.begin(); I != interior.end(); I++){
      RBF r(*I);
      rbfs.push_back(r);
    }
    for(I = boundary.begin(); I != boundary.end(); I++){
      RBF r(*I);
      rbfs.push_back(r);
    }
    
    //Now define the matrix to be inverted...
    matrix Mtemp(n+m,n+m);
    M = Mtemp;
    shared_ptr<const linear_diff_op2> L = thebvp -> get_linear_diff_op2(); 

    shared_ptr<const bdry_diff_op> B = thebvp -> get_bdry_diff_op();

    I = interior.begin();
    for(size_t i = 1; i <= n; i++){
      for(size_t j = 1; j <= n+m; j++)
	M(i,j) = L -> at(rbfs[j-1], *I);
      I++;
    }
    
    map<point, vector>::iterator J;
    J = normals.begin();
    for(size_t i = n+1; i <= n+m; i++){
      for(size_t j = 1; j <= n+m; j++)
	M(i,j) = B -> at(rbfs[j-1], J->first, J->second);
      J++;
    }
   
    computecoeffs();
    initted = true;
    rbfs_hash = hash_value(rbfs);
  }

  template<typename RBF>
  size_t interpolator<RBF>::hash_value(const std::vector<RBF>& rbfs_in)
  {
    size_t seed = 0;  

    for(size_t i = 0; i < rbfs_in.size(); i++)
      boost::hash_combine(seed,rbfs[i]);

    return seed;
  }
  

  //***** Evaluations, derivatives *************

  template<typename RBF>
  double interpolator<RBF>::operator()(const point& p) const{
    return at(p);
  }


  template<typename RBF>
  double interpolator<RBF>::at(const point& p) const{
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    std::vector<size_t> alpha; //empty vector
    diff_data loc = std::make_pair(p,alpha);
    if(remtable.find(loc) != remtable.end())
      return remtable[loc];
    
    double result = 0;
    for(size_t i = 1; i <= coeffs.size(); i++)
      result += coeffs(i)*rbfs[i-1].at(p);
    
    remtable[loc] = result;
    return result;
  }

  template<typename RBF>
  double interpolator<RBF>::d(const point& p, size_t k) const{
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    std::vector<size_t> alpha(k); alpha[k-1]++;
    diff_data loc = std::make_pair(p,alpha);
    if(remtable.find(loc) != remtable.end())
      return remtable[loc];

    double result = 0;
    for(size_t i = 1; i <= coeffs.size(); i++)
      result += coeffs(i)*rbfs[i-1].d(p,k);

    remtable[loc] = result;
    return result;
  }

  template<typename RBF>
  double interpolator<RBF>::d2(const point& p, size_t k1, size_t k2) const{
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    std::vector<size_t> alpha(k1>k2?k1:k2); alpha[k1-1]++; alpha[k2-1]++;
    diff_data loc = std::make_pair(p,alpha);
    if(remtable.find(loc) != remtable.end())
      return remtable[loc];

    double result = 0;
    for(size_t i = 1; i <= coeffs.size(); i++)
      result += coeffs(i)*rbfs[i-1].d2(p,k1,k2);
    return result;
  }

  //********** linear operations ***************

  template<typename RBF>
  interpolator<RBF> interpolator<RBF>::
  operator+(const interpolator<RBF>& u) const
  {    
    if(this -> rbfs_hash != u.rbfs_hash){
      badArgument exc;
      exc.reason = 
    	"Cannot add interpolators on different domains (interior and boundary must match)."; 
      exc.line = __LINE__;
      exc.file = __FILE__;
      throw exc;	
    }
    
    interpolator<RBF> out = *this;
    out.coeffs = (this -> coeffs) + u.coeffs;
   
    out.remtable.clear();
    
    return out;
  }

  template<typename RBF>
  interpolator<RBF> interpolator<RBF>::
  operator-(const interpolator<RBF>& u) const
  {    
    if(this -> rbfs_hash != u.rbfs_hash){
      badArgument exc;
      exc.reason = 
    	"Cannot subtract interpolators on different domains (interior and boundary must match)."; 
      exc.line = __LINE__;
      exc.file = __FILE__;
      throw exc;	
    }

    interpolator<RBF> out = *this;
    out.coeffs = (this -> coeffs) - u.coeffs;
    
    out.remtable.clear();
    
    return out;
  }
  
  template<typename RBF>
  interpolator<RBF> interpolator<RBF>::
  operator*(double a) const
  {
    interpolator<RBF> u = *this;
    u.coeffs = (this -> coeffs)*a;
    u.remtable.clear();
    return u;
  }

  template<typename RBF>
  interpolator<RBF> interpolator<RBF>::
  operator/(double a) const
  {
    interpolator<RBF> u = *this;
    u.coeffs = (this -> coeffs)*(1/a);
    u.remtable.clear();
    return u;
  }
  
  //********** data manipulation ***************

  template<typename RBF>
  void interpolator<RBF>::set_f(const realfunc& f){
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    thebvp -> set_f(f);
    computecoeffs();
  }

  template<typename RBF>
  void interpolator<RBF>::set_g(const realfunc& g){
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    thebvp -> set_g(g);
    computecoeffs();
  }

  template<typename RBF>
  void interpolator<RBF>::set_f(const map<point, double>& f){
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    thebvp -> set_f(f);
    computecoeffs();
  }

  template<typename RBF>
  void interpolator<RBF>::set_g(const map<point, double>& g){
    if(!initted){
      not_initted(__LINE__, __FILE__);
    }
    thebvp -> set_g(g);
    computecoeffs();
  }

  template<typename RBF>
  void interpolator<RBF>::not_initted(int line, string file) const{
    badArgument exc;
    exc.reason = 
      "Interpolator can't interpolate without initialisation data.";
    exc.line = line;
    exc.file = file;
    throw exc;
  }

  template<typename RBF>
  void interpolator<RBF>::computecoeffs(){
    using namespace std;
    linalg::vector rhs(n+m);

    map<point, double>::const_iterator I;

    I = (thebvp -> get_f()).begin();
    for(size_t i = 1; i <= n; i++){
      rhs(i) = I->second;
      I++;
    }
    I = (thebvp -> get_g()).begin();   
    for(size_t i = n+1; i <= n+m; i++){
      rhs(i) = I->second;
      I++;
    }

    coeffs = M.inv(rhs);
    remtable.clear();
  }

  //Instantiations
  using namespace rbf;
  template class interpolator<piecewise_polynomial>;
  template class interpolator<thin_plate_spline>;
  template class interpolator<multiquadric>;
  template class interpolator<inverse_multiquadric>;
  template class interpolator<inverse_quadratic>;
  template class interpolator<gaussian>;
}
